# config.tcl --
#
# Copyright (c) 2006 Georg Lehner <jorge@magma.com.ni>, 
#
# LEG03052006, See also http://www.magma.com.ni/moin/TclConfig
#
# config provides option managment for tcl programs.
#
# options can be provided via default values, configuration files,
# environment and the commandline, in this order.
#
# any number of option sets can be handled within one program
#
# minimal deployment:
#
#   package require config
#   ::config::add_options config <name> <description> \
#                                -default <value> -env <NAME>
#   ..
#   ::config::get_config config
#
# This will set $config(<name>) to one of:
#
#   <value>
#
#  the value of the variable <name> set in a tcl config file that
#    can be placed at various system wide or user specific places
#
#  the value of the environment variable <NAME>
#
#  the value of the -<n> switch on the commandline, where <n> is the
#    first letter of <name>
#
#
# config makes (optional) use of the cmdline package from tcllib
#
# Future directions:
#
# config is perceived as a framework for option managment with
# extensibility in mind.  Especially welcome would be:
#
#   - other parsers for configuration files
#   - writing optionsets to files
#   - other commandline parsing utilities
#   - a Tk-package for manipulating optionsets interactively
#   - manipulating optionsets on the console
#
# License:
#   This package is double licensed, you can either apply the
#   license Terms of the tcllib project: http://tcllib.sourceforge.net/
#   or the GNU General Public License Version 2 or newer:
#   http://www.gnu.org/licenses/gpl.html
#
# in short: Get it, use it, share it, improve it, but don't blame me.

namespace eval ::config {

set version 0.1

set optionset_help {

Description of optionsets:
--------------------------

In the following * means: not supported by cmdline package

An optionset is an array, describing options.  The optionset is first
defined, and can then be manipulated by searching the environment, the
commandline, several configuration or preferences files and user input
during the runtime of the program.

In the following list of reserved indexes 'option' containes the name
of an option

($option)
  holds the value of the option after option processing

(options): list of option names, auto generated by add_option
(shortcuts): list of the respective shortcuts
  No check is made to garantee unique shortcuts

($option,shortcut): token
  string:  the letter to use is the first letter of string
  true or unset:  use first letter of longname
  false:  do not define a short option *

($shortcut,option): option
  this reverse index of shortcuts to optionname is created during
  parsing of the options, put deleted afterwards

($option,secret): flag
  true:  do not show the option to the user
  false or unset:  show in usage messages, etc.
  when using cmdline this is made a boolean option

($option,default): defaultvalue
  not set:  option is a boolean
  else:  value to give the option if not defined elsewhere

($options,optional): flag
  only taken into account if defaultvalue exists
  true:  the option may be specified without an argument
         in which case it is set to the defaultvalue *
  not existent or false:  an argument to the option must be given

($option,description) "long description" *

($option,short) "(less then) one line description"
  this index is required

($option,env) name
  name of the asociated environment variable

(tags)  list of tags, each describing a configuration fileset
($tag,files) list of possible configuration files
($tag,parser) parser used for this configuration file
($tag,filename) first filename out of ($tag,files) which exists

If no tag is specified:
(files) list of files 
(filename) first filename out of (files)
(parser) parser to use for (filename)
}

set config_help {

The config package:
==================
global variables:

::config::optionsets  list of defined optionsets
::config::version  version of the config package

User procedures:
----------------

Define an optionset:

::config::add_option optionset name short_description ?options?
 options are:
 -secret
 -optional
 -default defaultvalue
 -shortcut shortcut
 -env envname

::config::add_file optionset ?parser? ?-tag tag? ?filename ..?
  Register a configuration file to be parsed with parser.
  Eventuall add a tag to distinguish configuration file setup.
  the list of files (filename ..) will be searched in order, the first file
  matched is used.

  If list of files is empty, search in several places with predefined
    names derived from argv0 and tag
  If tag is not specified set it to $argv0
  If parser is not specified set it to tcl


Get the options set:

::config::get_config optionset ?parser? ?usage? ?envname? ?shortcut?
  parse options from environment, config files and the commandline of
  a program.
  usage:  modfiy the usage message of cmdline
  shortcut:  modify the shortcut for the configfile option, default 'c'
  envname:  modify the environment variable used to look up the
            configfile, default CONFIG
  parser:  modify the parser used to parse the configfile, default tcl


::config::env_getoptions optionset
  set options from environment variables

::config::parse_file optionset ?tag?
  (eventually) set options by parsing the configuration file specified
  by tag, or the default configuration file if tag is not specified

::config::tcl_parse_file optionset filename
  parse filename in a safe interpreter and set optioneset with
  variables defined there.

::config::cmdline_getoptions argv optionset ?usage?
  parse the argumentstring in the variable argv into the optionset
  with the cmdline parser

::config::optionset optionset ?processor?
  optionset: name of an array holding optionset definition
  processor: name of the processor to use, by default cmdline

  The optionset is converted into a form usable by the processor and
  then return'ed

}


###: Optionsets

variable optionsets {}

proc add_option {optionsetname option short args} {
    
    upvar $optionsetname optionset
    
    set var ""
    foreach opt $args {
	if [string length $var] {
	    set optionset($option,$var) $opt
		set var ""
	    continue
	}
	
	switch -exact -- $opt {
	    -secret {set optionset($option,secret) true}
	    -optional {set optionset($option,optional) true}
	    
	    -default {set var default}
	    -description {set var description}
	    -shortcut {set var shortcut}
	    -env {set var env}

	    default {
		    error "unknown option $opt in $args"
	    }
	}
    }
    if [string length $var] {
	error "missing argument to -$var"
    }

    set optionset($option,short) $short
    set optionset($option) $optionset($option,default)
    lappend optionset(options) $option
}


###: Configuration files

proc add_file {optionsetname {parser tcl} args} {
    upvar $optionsetname optionset
    
    set tag_p [expr {[lindex $args 0] eq "-tag"}]
    if $tag_p {
	if {[llength $args] < 2} {error "missing tag"}
	set tag [lindex $args 1]
	set args [lrange $args 2 end]
    }
    
    # no filename specified, invent some names
    # $argv0.$parser config $argv0.conf $argv0.cfg $argv0
    # in several places
    # /etc /etc/$argv0 
    if ![llength $args] {
	package require cmdline
	if !$tag_p {
	    set basename [::cmdline::getArgv0]
	} else {
	    set basename $tag
	}
	# TODO: the heuristics for this
	# should be improved to allow
	# to specifiy a series of directories and extensions
	# and whirl them all together
	set args "\
                  /etc/$basename/$basename.$parser\
                  /etc/$basename/$basename.conf\
                  /etc/$basename/$basename.cfg\
                  /etc/$basename/config\
                  /etc/$basename.$parser\
                  /etc/$basename.conf\
                  /etc/$basename.cfg\
                  /etc/$basename\
                 "
	# Hack:
	# if tag is not specified, we most likely won't have
	# separate system and user preference processing
	if !$tag_p {
	    append args "\
                         ~/.$basename/$basename.$parser\
                         ~/.$basename/$basename.conf\
                         ~/.$basename/$basename.cfg\
                         ~/.${basename}rc
                        "
	}
    }
    if $tag_p {
	set optionset($tag,files) $args
	set optionset($tag,parser) $parser
	
	lappend optionset(tags) $tag
	
    } else {
	set optionset(files) $args
	set optionset(parser) $parser
	
	if ![info exists optionset(tags)] {
	    set optionset(tags) {}
	}
    }
    
}

proc _find_file {files} {

    foreach file [lappend files {}] {
	if [file readable $file] {
	    switch -exact [file type $file] {
		file {break}
		link {
		    set file [file readlink $file]
		    break
		}
	    }
	}
    }
    return $file
}

proc parse_file {optionsetname {tag {}}} {
    upvar $optionsetname optionset

    set tag_p [string length $tag]


    if $tag_p {
	set file_index ${tag},
	set filename_index ${tag},
	set parser_index ${tag},
    }
    append file_index files
    append filename_index filename
    append parser_index parser

    set filename [_find_file $optionset($file_index)]
    set optionset($filename_index) $filename
    set parser $optionset($parser_index)

    # TODO: maybe warn if no file found
    if ![string length $filename] {return}

    # execute the parser corresponding to the optionset
    ${parser}_parse_file optionset $filename
}


proc tcl_parse_file {optionsetname filename} {
    upvar $optionsetname optionset

    set parser [interp create -safe]
    interp invokehidden $parser source $filename

    foreach option $optionset(options) {
	if [interp eval $parser info exists $option] {
	    set optionset($option) [interp eval $parser set $option]
	}
    }

    interp delete $parser
}


###: Environment variable processing

proc env_getoptions {optionsetname} {
    upvar $optionsetname optionset

    foreach option $optionset(options) {
	if [info exists optionset($option,env)] {
	    set envname $optionset($option,env)
	    if [info exists ::env($envname)] {
		set optionset($option) $::env($envname)
	    }
	}
    }
}


###: Commandline procession

proc option_processor_config {optionset {processor cmdline}} {
    
    lappend optionsets $optionset
    
    switch -exact $processor {
	cmdline {return [_cmdline_optionset $optionset]}
	default {
	    error "wrong processor specified for optionset $optionset:\
	           $processor"
	}
    }
}


# tcllib cmdline processor

# see http://tcllib.sourceforge.net/doc/cmdline.html for a
# description of this commandline processor
proc _cmdline_optionset {optionsetname} {
    
    upvar 2 $optionsetname optionset
    set cmdline_def {}
    
    foreach option $optionset(options) {
	set option_dev {}
	set secret false
	
	# shortcut
	if [info exists optionset($option,shortcut)] {
	    
	    set shortcut $optionset($option,shortcut)
		if [string is boolean $shortcut] {
		    if $shortcut {set shortcut $option} else {continue}
		    # cmdline does not provide for long options
		    # so skip this option completely if shortcut is false
		    # TODO: -warn -error option?
		}
	} else {
	    set shortcut $option
	}
	set shortcut [string index $shortcut 0]
	set optionset($shortcut,option) $option
	lappend optionset(shortcuts) $shortcut

	set option_def $shortcut
	
	if [info exists optionset($option,secret)] {
	    set secret $optionset($option,secret)
	}
	
	if $secret {
	    append option_def .secret
	    # TODO: if a default value exists anyway, issue a
	    # warning 
	} else {
	    if [info exists optionset($option,default)] {
		append option_def .arg
		lappend option_def $optionset($option,default)
	    }
	}
	lappend option_def $optionset($option,short)
	lappend cmdline_def $option_def 
    }
    return $cmdline_def
}


# TODO: allow -known to be specified, invokes getKnownOptions
proc cmdline_getoptions {optionsetname argvname
			 {usage {}}
			 {strict false}} {
    package require cmdline

    upvar $optionsetname optset
    upvar $argvname arguments

    set opts [option_processor_config optset cmdline]

    # compile the getoptions command
    set getopts "::cmdline::get"

    append getopts [expr {$strict?"options":"KnownOptions"}]

    append getopts { arguments $opts}
    if [string length $usage] {append getopts { $usage}}

    # now evaluate it
    array set optset [eval $getopts]
    _cmdline_fix_shortcuts optset
}

# cmdline uses (forced by us) only one character option names so
# the value of ($option) is put into ($shortcut)
# this procedure fixes up an optionset, so that the options get the
# values of there shortcuts and the ($shortcut) entries get removed
proc _cmdline_fix_shortcuts {optionsetname} {
    upvar $optionsetname optionset

    foreach shortcut $optionset(shortcuts) {
	set option $optionset($shortcut,option)
	set optionset($option) $optionset($shortcut)
	unset optionset($shortcut)
	unset optionset($shortcut,option)
    }
}

###: Put it all together

# global optionset: config
#
proc get_config {optionsetname
		 {parser tcl}
		 {usage {}}
		 {envname CONFIG}
		 {shortcut c}} {

    upvar $optionsetname config

    array unset pre_config

    array set pre_config [array get config]
    
    add_file pre_config $parser

    if ![info exists pre_config(configfile,short)] {
	set secret_c_option true
	add_option pre_config configfile "path to configuration file" \
	    -shortcut $shortcut \
	    -env $envname \
	    -default $pre_config(files)\
	    -secret
    } else {set secret_c_option false}

    env_getoptions pre_config
    if [info exists pre_config(configfile)] {
	set pre_config(configfile,default) $pre_config(configfile)
    }

    set arg_list $::argv

    if {[catch {cmdline_getoptions pre_config arg_list $usage} retval]} {
	puts stderr $retval
	exit 1
    }


    # now we have the configfilename(s)
    set configfile $pre_config(configfile)

    
    add_option config configfile  "path to configuration file" \
	-shortcut $shortcut \
	-env $envname \
	-default $configfile

    if $secret_c_option {set config(configfile,secret) true}

    add_file config $parser $configfile
    set config(configfile) $configfile

    # parse the configfile, eventually the user config too
    parse_file config

    # fix up default options, so cmdline won't override what we have
    foreach option $config(options) {
	set config($option,default) $config($option)
    }


    # process the commandline arguments (strictly)
    if {[catch {cmdline_getoptions config ::argv $usage true} retval]} {
	puts stderr $retval
	exit 1
    }

    # now restore initial defaults
    foreach option $config(options) {
	set config($option,default) $pre_config($option,default)
    }
}

### Here ends the ::config namespace
}

package provide config $::config::version

if {[info ex argv0] && [file tail [info script]] == [file tail $argv0]} {
    pkg_mkIndex -verbose [file dirname [info script]]
}

# Thanks and Credits to:
# The Tcler's Wiki
# Richard Suchenwirth for the simple package example
# Arjen Markus for "safely source a file"
#